"""Auxiliary functions to connect to Project Gutenberg's library"""

from io import BytesIO
from typing import Final, Optional

import requests

from gutenberg2kindle.config import (
    FORMAT_AUTO,
    FORMAT_IMAGES,
    FORMAT_NO_IMAGES,
    SETTINGS_FORMAT,
    get_config,
)

GUTENBERG_BOOK_WITH_IMAGES_BASE_URL: Final[str] = (
    "https://www.gutenberg.org/ebooks/{book_id}.kindle.images"
)
GUTENBERG_BOOK_BASE_URL: Final[str] = (
    "https://www.gutenberg.org/ebooks/{book_id}.kindle"
)


def download_book(book_id: int) -> Optional[BytesIO]:
    """
    Given a Gutenberg book ID as an integer, and the expected format,
    fetches the content of the book into memory and returns it wrapped
    in a Bytes IO instance.
    """

    fmt = get_config(SETTINGS_FORMAT)
    assert isinstance(fmt, str)

    book_url = GUTENBERG_BOOK_BASE_URL.format(book_id=book_id)
    if fmt == FORMAT_NO_IMAGES:
        return fetch_book_from_url(book_url)

    book_with_images_url = GUTENBERG_BOOK_WITH_IMAGES_BASE_URL.format(
        book_id=book_id
    )
    if fmt == FORMAT_IMAGES:
        return fetch_book_from_url(book_with_images_url)

    if fmt == FORMAT_AUTO:
        book_or_none = fetch_book_from_url(book_with_images_url)

        if book_or_none is not None:
            return book_or_none

        return fetch_book_from_url(book_url)

    raise ValueError(f"{fmt} is an invalid format")


def fetch_book_from_url(book_url: str) -> Optional[BytesIO]:
    """
    Given a Gutenberg book URL, fetches the content
    of the book into memory and returns it wrapped in a BytesIO
    instance
    """

    response = requests.get(book_url)
    if response.status_code != 200:
        return None
    book_content = response.content

    memory_bytes = BytesIO()
    memory_bytes.write(book_content)
    memory_bytes.seek(0)
    return memory_bytes



"""
Functions and other helpers for `gutenberg2kindle`'s command-line interface.
"""

import argparse
import getpass
import socket
import sys
from typing import Final, Optional, Union

from gutenberg2kindle.config import (
    AVAILABLE_SETTINGS,
    get_config,
    set_config,
    setup_settings,
)
from gutenberg2kindle.email import send_book
from gutenberg2kindle.gutenberg import download_book

COMMAND_SEND: Final[str] = "send"
COMMAND_GET_CONFIG: Final[str] = "get-config"
COMMAND_SET_CONFIG: Final[str] = "set-config"
AVAILABLE_COMMANDS: Final[list[str]] = [
    COMMAND_SEND,
    COMMAND_GET_CONFIG,
    COMMAND_SET_CONFIG,
]


def get_parser() -> argparse.ArgumentParser:
    """
    Generate and return a parser for the CLI tool
    """
    parser = argparse.ArgumentParser(
        prog="gutenberg2kindle",
        description=(
            "A CLI tool to download and send ebooks "
            "from Project Gutenberg to a Kindle email "
            "address via SMTP"
        ),
        epilog="Happy reading! :-)"
    )
    parser.add_argument(
        "command",
        metavar="COMMAND",
        type=str,
        choices=AVAILABLE_COMMANDS,
        help=(
            "Command to use. Supported options allow the user to "
            "either set the tool's config options, read the current "
            "config, or send some books using the current config."
        ),
    )
    parser.add_argument(
        "--book-id",
        "-b",
        metavar="BOOK_ID",
        type=int,
        nargs='+',
        default=[],
        help=(
            "ID of the Project Gutenberg book you want to download. "
            "Can be specified multiple times to download more than one "
            "book at the same time."
        ),
    )
    parser.add_argument(
        "--name",
        "-n",
        metavar="NAME",
        type=str,
        choices=AVAILABLE_SETTINGS,
        help=(
            "Setting to get / set, whenever these commands are used."
        ),
    )
    parser.add_argument(
        "--value",
        "-a",
        metavar="VALUE",
        type=str,
        help=(
            "Value to set for the specified setting name, if required."
        ),
    )

    return parser


def format_setting(name: str, value: Union[str, int]) -> str:
    """Formats a setting with its value for printing"""
    return f"{name}:\t\t{value}"


def main() -> None:
    """
    Run the tool's CLI
    """

    # setup tool
    setup_settings()

    # parse arguments
    parser = get_parser()
    args = parser.parse_args()

    # cast certain arguments to expected types
    command: str = args.command
    name: Optional[str] = args.name
    value: Optional[str] = args.value

    if command == "send":
        book_id_list: list[int] = args.book_id
        books_amount = len(book_id_list)

        # request password
        password = getpass.getpass("Please enter your SMTP password: ")

        for book_id in book_id_list:
            book = download_book(book_id)

            if book is None:
                print(f"Book `{book_id}` could not be downloaded!")
                sys.exit(1)

            print(f"Sending book `{book_id}`...")
            try:
                send_book(book_id, book, password)
                book.close()
            except socket.error as err:  # pylint: disable=no-member
                print(
                    "SMTP credentials are invalid! "
                    "Please validate your current config.\n"
                    f"Server error message: {err}"
                )
                sys.exit(1)
            print(f"Book `{book_id}` sent!")

        if books_amount > 1:
            print(f"{books_amount} books sent successfully!")

    elif command == "get-config":
        stored_value = get_config(name)

        if isinstance(stored_value, dict):
            for setting_name, setting_value in stored_value.items():
                print(format_setting(setting_name, setting_value))
        else:
            print(stored_value)

    elif command == "set-config":
        if name is None:
            print("Please specify a setting name with the `--name` flag")
            sys.exit(1)

        if value is None:
            print("Please specify a setting value with the `--value` flag")
            sys.exit(1)

        set_config(name, value)
        print(format_setting(name, value))


if __name__ == "__main__":
    main()



"""Auxiliary module with functions that help with sending email"""

import smtplib
import ssl
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from io import BytesIO
from typing import Final

from gutenberg2kindle.config import (
    SETTINGS_KINDLE_EMAIL,
    SETTINGS_SENDER_EMAIL,
    SETTINGS_SMTP_PORT,
    SETTINGS_SMTP_SERVER,
    get_config,
)

EMAIL_SUBJECT: Final[str] = "Your Project Gutenberg ebook!"
EMAIL_BODY: Final[str] = "- Sent with gutenberg2kindle. Happy reading!"


def create_base_email(
    sender_email: str,
    kindle_email: str,
) -> MIMEMultipart:
    """Generates and returns the base email to use when sending a book"""

    message = MIMEMultipart()
    message["From"] = sender_email
    message["To"] = kindle_email
    message["Subject"] = EMAIL_SUBJECT
    message.attach(MIMEText(EMAIL_BODY, "plain"))
    return message


def send_book(
    book_id: int,
    book_in_memory: BytesIO,
    password: str,
) -> None:
    """
    Given a book as a file in memory, sends the file via email
    using the stored config
    """

    # retrieving config
    sender_email = get_config(SETTINGS_SENDER_EMAIL)
    assert isinstance(sender_email, str)

    kindle_email = get_config(SETTINGS_KINDLE_EMAIL)
    assert isinstance(kindle_email, str)

    smtp_server = get_config(SETTINGS_SMTP_SERVER)
    assert isinstance(smtp_server, str)

    port = get_config(SETTINGS_SMTP_PORT)
    assert isinstance(port, int)

    # creating email message
    message = create_base_email(sender_email, kindle_email)

    # loading the file as an attachment
    part = MIMEBase("application", "octet-stream")
    part.set_payload(book_in_memory.read())
    encoders.encode_base64(part)
    part.add_header(
        "Content-Disposition",
        f"attachment; filename={book_id}.mobi",
    )
    message.attach(part)

    # send email
    text = message.as_string()
    context = ssl.create_default_context()
    with smtplib.SMTP(smtp_server, port) as server:
        server.starttls(context=context)
        server.login(sender_email, password)
        server.sendmail(sender_email, kindle_email, text)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gutenberg2kindle']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.26.0,<3.0.0', 'usersettings>=1.1.5,<2.0.0']

entry_points = \
{'console_scripts': ['gutenberg2kindle = gutenberg2kindle.cli:main']}

setup_kwargs = {
    'name': 'gutenberg2kindle',
    'version': '0.2.1',
    'description': 'A small Python tool to download and send ebooks from Project Gutenberg to a Kindle email address via SMTP',
    'long_description': "# Gutenberg2Kindle\n\nA small Python tool to download and send ebooks from Project Gutenberg to a Kindle email address via SMTP\n\n## What's this?\n\n`gutenberg2kindle` is a small command-line interface tool that aims to automatically download a `.mobi` book from [Project Gutenberg](https://www.gutenberg.org/)'s library of free books in the public domain, and then send the ebook's file to a Kindle email address (although, generally, it can be sent to any email address), with just one command.\n\nThe book is sent through a SMTP server with TLS, requiring the user to configure the server settings beforehand via tool commands.\n\n## Installation\n\nYou can use your Python package manager (e.g. [pip](https://pip.pypa.io/en/stable/)) to install `gutenberg2kindle`.\n\n```bash\npip install gutenberg2kindle\n```\n\n## Usage\n\n`gutenberg2kindle` comes with a command-line interface; its help text can be accessed via:\n\n```bash\ngutenberg2kindle --help\n```\n\nYou can check the tool's current configuration via:\n\n```bash\n# will print all config variables with their current values\ngutenberg2kindle get-config\n\n# will print only the value for the key you're specifying\ngutenberg2kindle get-config --name <key name>\n```\n\nYou can set a value for any of the settings via:\n\n```bash\ngutenberg2kindle set-config --name <key name> --value <key value>\n```\n\nFinally, once you're done configuring your project, you can send any ebook via its Project Gutenberg book ID (with flags `-b` or `--book-id`):\n\n```bash\ngutenberg2kindle send -b <book id as an integer, e.g. 1>\n```\n\nYou can send multiple books at the same time in the same run, the `-b` / `--book-id` flag accepts multiple arguments.\n\n```bash\ngutenberg2kindle send -b <first book id> [<second book id> <third book id>...]\n```\n\nNote that, if using Gmail as your SMTP server, you might need to set up an [App Password](https://support.google.com/accounts/answer/185833) to use instead of your regular password.\n\n## Contributing\n\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change. Contributions for issues that are already open by maintainers are welcome and encouraged.\n\nPlease make sure to update tests as appropriate; a minimum coverage of 75% is expected (and enforced by Github Actions!).\n\n## License\n\nThis project is licensed under the [GNU Affero General Public License v3.0](https://github.com/aitorres/gutenberg2kindle/blob/main/LICENSE).\n",
    'author': 'Andrés Ignacio Torres',
    'author_email': 'dev@aitorres.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aitorres/gutenberg2kindle',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""
Auxiliary functions to handle user configuration for `gutenberg2kindle`
"""

from typing import Final, Optional, Union

import usersettings  # type: ignore


SETTINGS_SMTP_SERVER: Final[str] = "smtp_server"
SETTINGS_SMTP_PORT: Final[str] = "smtp_port"
SETTINGS_SENDER_EMAIL: Final[str] = "sender_email"
SETTINGS_KINDLE_EMAIL: Final[str] = "kindle_email"
SETTINGS_FORMAT: Final[str] = "format"
AVAILABLE_SETTINGS: Final[list[str]] = [
    SETTINGS_SMTP_SERVER,
    SETTINGS_SMTP_PORT,
    SETTINGS_SENDER_EMAIL,
    SETTINGS_KINDLE_EMAIL,
    SETTINGS_FORMAT,
]

FORMAT_IMAGES: Final[str] = "images"
FORMAT_NO_IMAGES: Final[str] = "no_images"
FORMAT_AUTO: Final[str] = "auto"
VALID_FORMATS: Final[list[str]] = [
    FORMAT_IMAGES,
    FORMAT_NO_IMAGES,
    FORMAT_AUTO,
]

settings = usersettings.Settings("gutenberg2kindle")


def setup_settings() -> usersettings.Settings:
    """
    Sets up and returns an instance of the `usersettings.Settings` model
    with values for all the required settings used in the project
    """

    settings.add_setting(SETTINGS_SMTP_SERVER, str, "")
    settings.add_setting(SETTINGS_SMTP_PORT, int, 465)
    settings.add_setting(SETTINGS_SENDER_EMAIL, str, "")
    settings.add_setting(SETTINGS_KINDLE_EMAIL, str, "")
    settings.add_setting(SETTINGS_FORMAT, str, FORMAT_AUTO)
    settings.load_settings()


def get_config(name: Optional[str] = None) -> Optional[Union[dict, int, str]]:
    """
    Given a setting name, returns the value for said setting.

    If no name is given, returns a dictionary with all settings
    and their corresponding values.
    """

    if name is None:
        return settings

    if name not in AVAILABLE_SETTINGS:
        raise ValueError(f"`{name}` is not a valid setting name")

    return settings[name]


def set_config(name: str, value: Union[int, str]) -> None:
    """
    Given a setting name and a value, sets the value for said setting.
    """

    if name not in AVAILABLE_SETTINGS:
        raise ValueError(f"`{name}` is not a valid setting name")

    if name == SETTINGS_FORMAT and value not in VALID_FORMATS:
        raise ValueError(
            f"`{value}` is not a valid format "
            f"(expected one of: {VALID_FORMATS})"
        )

    settings[name] = value
    settings.save_settings()
